"""
Test for RateML module

.. moduleauthor:: Aaron Perez Martin <a.perez.martin@fz-juelich.de>
.. moduleauthor:: Michiel van der Vlag <m.van.der.vlag@fz-juelich.de>

"""
import importlib

import pytest, os, itertools, numpy as np, re, sys
from tvb.rateML import XML2model
from tvb.rateML.XML2model import RateML
from tvb.simulator.models.base import Model
pycuda = pytest.importorskip("pycuda")
try:
	import pycuda.autoinit
	PYCUDA_OK = True
except ImportError:
	PYCUDA_OK = False
skip_cuda_if_not_avail = pytest.mark.skipif(
	not PYCUDA_OK, reason='import pycuda.autoinit failed, CUDA not available')

xmlModelTesting = "kuramoto.xml"
framework_path, _ = os.path.split(XML2model.__file__)
XMLModel_path = os.path.join(framework_path, "XMLmodels")
generatedModels_path = os.path.join(framework_path, "generatedModels")
cuda_ref_path = os.path.join(generatedModels_path, "cuda_refs")
run_path = os.path.join(framework_path, "run")
dic_regex_mincount = {r'^__global':1,
					  r'state\(\(\(':1,
					  r'state\(\(t':2,
					  r'tavg\(':1,
					  r'= params\(\d\)\;$':2}
def compiler_opts():
	opts = ['--ptxas-options=-v', '-maxrregcount=32', '-lineinfo']
	opts.append('-lineinfo')
	opts.append('-DWARP_SIZE=%d' % (32,))
	opts.append('-DBLOCK_DIM_X=%d' % (32,))
	opts.append('-DNH=%s' % ('nh',))

def compile_cuda_model(location, model_name):
	import pycuda.autoinit
	from pycuda.compiler import SourceModule

	source_file = os.path.join(location, model_name + ".c")
	compiled = False
	with open(source_file, 'r') as f:
		mod_content = f.read().replace('M_PI_F', '%ff' % (np.pi,))
		# TODO replace lower pi and inf need to be fixed to whole words only
		# mod_content = mod_content.replace('pi', '%ff' % (np.pi,))
		# mod_content = mod_content.replace('inf', 'INFINITY')

		# Compile model
		mod = SourceModule(mod_content, options=compiler_opts(), include_dirs=[], no_extern_c=True, keep=False)
		if mod is not None:
			compiled = True
	return compiled

def check_input_params():
	if len(sys.argv)>=2:
		sys.argv = sys.argv[1:]

def find_attributes(obj, pattern):
	dict_found = {}
	count = 0
	for k,v in obj.__dict__.items():
		if k.startswith(pattern):
			dict_found[k] = v
			count +=1
	return dict_found, count

def remove_model_driver(model):

	rm_model = generatedModels_path + '/' + model + '.c'
	rm_driver = run_path + '/model_driver.py'
	if os.path.exists(rm_model):
		os.remove(rm_model)
	if os.path.exists(rm_driver):
		os.remove(rm_driver)

def setup_namespace(model='kuramoto'):

	# remove files to be sure there is always a new version
	remove_model_driver(model)

	# gemerate model and setup namespace for every test
	RateML(model, 'cuda')
	driver = importlib.import_module('.model_driver_' + model, 'tvb.rateML.run')
	driver = driver.Driver_Execute(driver.Driver_Setup())

	return driver

#This line avoid passing script name as a parameters
check_input_params()
class TestRateML():
	models=["epileptor", "kuramoto", "montbrio", "oscillator", "rwongwang"]
	python_mods = ["python"]*len(models)
	languages = ["python", "cuda"]

	# Cuda Section
	#----------------
	@skip_cuda_if_not_avail
	def test_make_cuda_setup(self):

		driver = setup_namespace('oscillator')
		bx, by = driver.args.blockszx, driver.args.blockszy
		nwi = driver.n_work_items
		rootnwi = int(np.ceil(np.sqrt(nwi)))
		gridx = int(np.ceil(rootnwi / bx))
		gridy = int(np.ceil(rootnwi / by))

		assert gridx * gridy * bx * by >= nwi

	@skip_cuda_if_not_avail
	def test_make_cuda_data(self):
		data = {}
		data["serie"] = np.zeros(2, 'f')

		driver = setup_namespace('kuramoto')
		gpu_data = driver.make_gpu_data(data)

		assert gpu_data["serie"].size == data["serie"].size

	@skip_cuda_if_not_avail
	def test_make_cuda_kernel(self):
		driver = setup_namespace()
		step_fn = driver.make_kernel(source_file=driver.args.filename, warp_size=32, args=driver.args,
						   lineinfo=driver.args.lineinfo, nh=driver.buf_len)
		assert step_fn is not None and len(str(step_fn))>0

	# Model Driver Section
	# --------------------
	def test_check_parameters(self):

		# works for default workitems settings
		driver = setup_namespace('oscillator')
		_, count = find_attributes(driver.args, "n_sweep_")
		assert count == 2
		assert driver.exposures == 2
		assert driver.states == 2

		# 16 workitems is based on driver default
		n_work_items, n_params = driver.params.shape
		assert n_work_items == 16 and n_params == 2

	# Model Section
	# ----------------
	@pytest.mark.slow
	@pytest.mark.parametrize('model_name', models)
	def test_load_model(self, model_name):
		model, _, _, _, _ = RateML(model_name).load_model()
		assert model is not None

	@pytest.mark.slow
	def test_prep_model_bound(self):
		# python only
		_, svboundaries, _, _, _ = RateML("kuramoto", language="cuda").load_model()
		assert svboundaries

	@pytest.mark.slow
	def test_prep_model_noise(self):
		# works for xml models which have noise enabled and nsig defined.
		# cuda only
		_, _, _, noise, nsig = RateML('kuramoto', 'cuda').load_model()
		assert noise
		assert nsig

	@pytest.mark.slow
	def test_prep_model_coupling(self):
		# cuda only
		_, _, couplinglist, _, _ = RateML("oscillator", "cuda").load_model()
		assert len(couplinglist) > 0

	@pytest.mark.slow
	def test_time_serie(self):
		driver = setup_namespace('oscillator')
		driver.args.n_time = 150
		tavg0 = driver.run_simulation()
		assert np.allclose(driver.compare_with_ref(tavg0), 1, 1e-6, 1e-6)

	@pytest.mark.slow
	@pytest.mark.parametrize('model_name, language', itertools.product(models, languages))
	def test_convert_model(self, model_name, language):
		if language == 'cuda' and not PYCUDA_OK:
			return
		# reload(Driver_Execute)
		model_str, driver_str = RateML(model_filename=model_name,language=language,XMLfolder=XMLModel_path,
									   GENfolder=generatedModels_path).render()
		if language == "python":
			assert len(model_str) >0 and driver_str is None
		if language == "cuda":
			assert len(model_str) >0 and driver_str is not None and len(driver_str) > 0

	@skip_cuda_if_not_avail
	@pytest.mark.slow
	@pytest.mark.parametrize('model_name', models)
	def test_compile_cuda_models(self, model_name):
		assert compile_cuda_model(location=generatedModels_path, model_name=model_name)

	@skip_cuda_if_not_avail
	@pytest.mark.slow
	@pytest.mark.parametrize('model_name', models)
	def test_contentcheck_cuda_models(self, model_name):

		source_file = os.path.join(generatedModels_path, model_name + ".c")
		with open(source_file, "r") as f:
			lines = f.read()

			# trucate the file to avoid processing bold_update
			lines = lines.split("__global__ void bold_update")[0]

			pattern_model = r'^__global__ void ' + model_name + '\('
			assert len(re.findall(pattern=pattern_model, string=lines, flags=re.IGNORECASE + re.MULTILINE + re.DOTALL)) >0

			for regex, mincount in dic_regex_mincount.items():
				matches = re.findall(pattern=regex, string=lines, flags=re.IGNORECASE + re.MULTILINE + re.DOTALL)
				assert len(matches) >= dic_regex_mincount[regex]

	# Simulation Section
	# ----------------
	@skip_cuda_if_not_avail
	@pytest.mark.slow
	def test_simulation_cuda_model_osc(self):
		# simulate with different properties to check if output shape is related

		driver = setup_namespace('oscillator')
		driver.args.n_time = 4
		driver.exposures = 2
		driver.n_regions = 68
		driver.n_work_items = 8
		tavg_data = driver.run_simulation()

		# a = n_steps, b = n_expos, c = n_regions, d = n_workitems
		a, b, c, d = tavg_data.shape
		assert (a, b, c, d) == (4, 2, 68, 8)

	@skip_cuda_if_not_avail
	@pytest.mark.slow
	def test_simulation_cuda_model_kur(self):
		# simulate with different properties to check if output shape is related

		driver = setup_namespace('kuramoto')
		driver.args.n_time = 8
		driver.exposures = 1
		driver.args.n_regions = 76
		driver.n_work_items = 12
		tavg_data = driver.run_simulation()

		# a = n_steps, b = n_expos, c = n_regions, d = n_workitems
		a, b, c, d = tavg_data.shape
		assert (a, b, c, d) == (8, 1, 76, 12)

	@skip_cuda_if_not_avail
	@pytest.mark.slow
	def test_simulation_cuda_model_epi(self):
		# simulate with different properties to check if output shape is related

		driver = setup_namespace('epileptor')
		driver.args.n_time = 12
		driver.exposures = 6
		driver.args.n_regions = 96
		driver.n_work_items = 18
		tavg_data = driver.run_simulation()

		# a = n_steps, b = n_expos, c = n_regions, d = n_workitems
		a, b, c, d = tavg_data.shape
		assert (a, b, c, d) == (12, 6, 96, 18)

	def test_load_model(self):
		name = 'epileptor'
		model, _, _, _, _ = RateML(name).load_model()
		assert model is not None

	def test_render_model(self):
		name = 'epileptor'
		model_str, _ = RateML(name).render()
		assert '_numba_dfun_EpileptorT' in model_str

	# TODO fix
	# def test_eval_model_str(self):
	# 	filename = 'epileptor'
	# 	classname = 'EpileptorT'
	# 	module = {}
	# 	exec(RateML(filename).render()[0], module)
	# 	assert issubclass(module[classname], Model)
	# 	model = module[classname]()
	# 	assert isinstance(model, Model)




from __future__ import print_function

import logging
import itertools
import argparse
import pickle

from tvb.simulator.lab import *
from tvb.basic.logger.builder import get_logger

from tvb.rateML.run.regular_run import regularRun

import os.path
import numpy as np
import pycuda.autoinit
import pycuda.driver as drv
from pycuda.compiler import SourceModule
import pycuda.gpuarray as gpuarray

import matplotlib.pyplot as plt

import time
import tqdm

here = os.path.dirname(os.path.abspath(__file__))

class Driver_Setup:

	def __init__(self):
		self.args = self.parse_args()

		self.logger = get_logger('tvb.rateML')
		self.logger.setLevel(level='INFO' if self.args.verbose else 'WARNING')

		self.checkargbounds()

		self.dt = self.args.delta_time
		self.connectivity = self.tvb_connectivity(self.args.n_regions)
		self.weights = self.connectivity.weights
		self.lengths = self.connectivity.tract_lengths
		self.tavg_period = 1.0
		self.n_inner_steps = int(self.tavg_period / self.dt)

		self.params = self.setup_params(
		self.args.n_sweep_arg0,
		self.args.n_sweep_arg1,
		)

		# bufferlength is based on the minimum of the first swept parameter (speed for many tvb models)
		self.n_work_items, self.n_params = self.params.shape
		self.buf_len_ = ((self.lengths / self.args.speeds_min / self.dt).astype('i').max() + 1)
		self.buf_len = 2 ** np.argwhere(2 ** np.r_[:30] > self.buf_len_)[0][0]  # use next power of

		self.states = self.args.states
		self.exposures = self.args.exposures

		if self.args.gpu_info:
			self.logger.setLevel(level='INFO')
			self.gpu_device_info()
			exit(1)

		self.logdata()

	def logdata(self):

		self.logger.info('dt %f', self.dt)
		self.logger.info('n_nodes %d', self.args.n_regions)
		self.logger.info('weights.shape %s', self.weights.shape)
		self.logger.info('lengths.shape %s', self.lengths.shape)
		self.logger.info('tavg period %s', self.tavg_period)
		self.logger.info('n_inner_steps %s', self.n_inner_steps)
		self.logger.info('params shape %s', self.params.shape)

		self.logger.info('nstep %d', self.args.n_time)
		self.logger.info('n_inner_steps %f', self.n_inner_steps)

		# self.logger.info('single connectome, %d x %d parameter space', self.args.n_sweep_arg0, self.args.n_sweep_arg1)
		self.logger.info('real buf_len %d, using power of 2 %d', self.buf_len_, self.buf_len)
		self.logger.info('number of states %d', self.states)
		self.logger.info('model %s', self.args.model)
		self.logger.info('real buf_len %d, using power of 2 %d', self.buf_len_, self.buf_len)
		self.logger.info('memory for states array on GPU %d MiB',
						 (self.buf_len * self.n_work_items * self.states * self.args.n_regions * 4) / 1024 ** 2)

	def checkargbounds(self):

		try:
			assert self.args.n_sweep_arg0 > 0, "Min value for [N_SWEEP_ARG0] is 1"
			assert self.args.n_time > 0, "Minimum number for [-n N_TIME] is 1"
			assert self.args.n_regions > 0, "Min value for  [-tvbn n_regions] for default data set is 68"
			assert self.args.blockszx > 0 and self.args.blockszx <= 32,	"Bounds for [-bx BLOCKSZX] are 0 < value <= 32"
			assert self.args.blockszy > 0 and self.args.blockszy <= 32, "Bounds for [-by BLOCKSZY] are 0 < value <= 32"
			assert self.args.delta_time > 0.0, "Min value for [-dt delta_time] is > 0.0, default is 0.1"
			assert self.args.speeds_min > 0.0, "Min value for [-sm speeds_min] is > 0.0, default is 3e-3"
			assert self.args.exposures > 0, "Min value for [-x exposures] is 1"
			assert self.args.states > 0, "Min value for [-s states] is 1"
		except AssertionError as e:
			self.logger.error('%s', e)
			raise

	def tvb_connectivity(self, tvbnodes):
		# white_matter = connectivity.Connectivity.from_file(source_file="connectivity_"+str(tvbnodes)+".zip")
		white_matter = connectivity.Connectivity.from_file(source_file="paupau.zip")
		white_matter.configure()
		return white_matter

	def parse_args(self):  # {{{
		parser = argparse.ArgumentParser(description='Run parameter sweep.')

		# for every parameter that needs to be swept, the size can be set
		parser.add_argument('-s0', '--n_sweep_arg0', default=4, help='num grid points for 1st parameter', type=int)
		parser.add_argument('-s1', '--n_sweep_arg1', default=4, help='num grid points for 2st parameter', type=int)
		parser.add_argument('-n', '--n_time', default=400, help='number of time steps to do', type=int)
		parser.add_argument('-v', '--verbose', default=False, help='increase logging verbosity', action='store_true')
		parser.add_argument('-m', '--model', default='montbrio', help="neural mass model to be used during the simulation")
		parser.add_argument('-s', '--states', default=2, type=int, help="number of states for model")
		parser.add_argument('-x', '--exposures', default=2, type=int, help="number of exposures for model")
		parser.add_argument('-l', '--lineinfo', default=False, help='generate line-number information for device code.', action='store_true')
		parser.add_argument('-bx', '--blockszx', default=8, type=int, help="gpu block size x")
		parser.add_argument('-by', '--blockszy', default=8, type=int, help="gpu block size y")
		parser.add_argument('-val', '--validate', default=False, help="enable validation with refmodels", action='store_true')
		parser.add_argument('-r', '--n_regions', default="68", type=int, help="number of tvb nodes")
		parser.add_argument('-p', '--plot_data', type=int, help="plot res data for selected state")
		parser.add_argument('-w', '--write_data', default=False, help="write output data to file: 'tavg_data", action='store_true')
		parser.add_argument('-g', '--gpu_info', default=False, help="show gpu info", action='store_true')
		parser.add_argument('-dt', '--delta_time', default=0.1, type=float, help="dt for simulation")
		parser.add_argument('-sm', '--speeds_min', default=1, type=float, help="min speed for temporal buffer")

		args = parser.parse_args()
		return args


	def setup_params(self,
		n0,
		n1,
		):
		'''
		This code generates the parameters ranges that need to be set
		'''
		sweeparam0 = np.linspace(1.0, 1.0, n0)
		sweeparam1 = np.linspace(1.0, 1.0, n1)
		params = itertools.product(
		sweeparam0,
		sweeparam1,
		)
		params = np.array([vals for vals in params], np.float32)
		return params


	def gpu_device_info(self):
		'''
		Get GPU device information
		TODO use this information to give user GPU setting suggestions
		'''
		dev = drv.Device(0)
		print('\n')
		self.logger.info('GPU = %s', dev.name())
		self.logger.info('TOTAL AVAIL MEMORY: %d MiB', dev.total_memory()/1024/1024)

		# get device information
		att = {'MAX_THREADS_PER_BLOCK': [],
			   'MAX_BLOCK_DIM_X': [],
			   'MAX_BLOCK_DIM_Y': [],
			   'MAX_BLOCK_DIM_Z': [],
			   'MAX_GRID_DIM_X': [],
			   'MAX_GRID_DIM_Y': [],
			   'MAX_GRID_DIM_Z': [],
			   'TOTAL_CONSTANT_MEMORY': [],
			   'WARP_SIZE': [],
			   # 'MAX_PITCH': [],
			   'CLOCK_RATE': [],
			   'TEXTURE_ALIGNMENT': [],
			   # 'GPU_OVERLAP': [],
			   'MULTIPROCESSOR_COUNT': [],
			   'SHARED_MEMORY_PER_BLOCK': [],
			   'MAX_SHARED_MEMORY_PER_BLOCK': [],
			   'REGISTERS_PER_BLOCK': [],
			   'MAX_REGISTERS_PER_BLOCK': []}

		for key in att:
			getstring = 'drv.device_attribute.' + key
			# att[key].append(eval(getstring))
			self.logger.info(key + ': %s', dev.get_attribute(eval(getstring)))

class Driver_Execute(Driver_Setup):

	def __init__(self, ds):
		self.args = ds.args
		self.set_CUDAmodel_dir()
		self.weights, self.lengths, self.params = ds.weights, ds.lengths, ds.params
		self.buf_len, self.states, self.n_work_items = ds.buf_len, ds.states, ds.n_work_items
		self.n_inner_steps, self.n_params, self.dt = ds.n_inner_steps, ds.n_params, ds.dt
		self.exposures, self.logger = ds.exposures, ds.logger

	def set_CUDAmodel_dir(self):
		self.args.filename = os.path.join((os.path.dirname(os.path.abspath(__file__))),
								 "../generatedModels", self.args.model.lower() + '.c')

	def set_CUDA_ref_model_dir(self):
		self.args.filename = os.path.join((os.path.dirname(os.path.abspath(__file__))),
								 "../generatedModels/cuda_refs", self.args.model.lower() + '.c')

	def compare_with_ref(self, tavg0):
		self.args.model = self.args.model + 'ref'
		self.set_CUDA_ref_model_dir()
		tavg1 = self.run_simulation()

		# compare output to check if same as template
		# comparison = (tavg0.ravel() == tavg1.ravel())
		# self.logger.info('Templated version is similar to original %d:', comparison.all())
		self.logger.info('Corr.coef. of model with %s is: %f', self.args.model,
						 np.corrcoef(tavg0.ravel(), tavg1.ravel())[0, 1])

		return np.corrcoef(tavg0.ravel(), tavg1.ravel())[0, 1]

	def make_kernel(self, source_file, warp_size, args, lineinfo=False, nh='nh'):

		try:
			with open(source_file, 'r') as fd:
				source = fd.read()
				source = source.replace('pi', '%ff' % (np.pi, ))
				source = source.replace('inf', 'INFINITY')
				opts = ['--ptxas-options=-v', '-maxrregcount=32']
				if lineinfo:
					opts.append('-lineinfo')
				opts.append('-DWARP_SIZE=%d' % (warp_size, ))
				opts.append('-DNH=%s' % (nh, ))

				idirs = [here]
				self.logger.info('nvcc options %r', opts)

				try:
					network_module = SourceModule(
							source, options=opts, include_dirs=idirs,
							no_extern_c=True,
							keep=False,)
				except drv.CompileError as e:
					self.logger.error('Compilation failure \n %s', e)
					exit(1)

				# generic func signature creation
				mod_func = "{}{}{}{}".format('_Z', len(args.model), args.model, 'jjjjjfPfS_S_S_S_')

				step_fn = network_module.get_function(mod_func)

		except FileNotFoundError as e:
			self.logger.error('%s.\n  Generated model filename should match model on cmdline', e)
			exit(1)

		return step_fn #}}}

	def cf(self, array):#{{{
		# coerce possibly mixed-stride, double precision array to C-order single precision
		return array.astype(dtype='f', order='C', copy=True)#}}}

	def nbytes(self, data):#{{{
		# count total bytes used in all data arrays
		nbytes = 0
		for name, array in data.items():
			nbytes += array.nbytes
		return nbytes#}}}

	def make_gpu_data(self, data):#{{{
		# put data onto gpu
		gpu_data = {}
		for name, array in data.items():
			try:
				gpu_data[name] = gpuarray.to_gpu(self.cf(array))
			except drv.MemoryError as e:
				self.gpu_mem_info()
				self.logger.error(
					'%s.\n\t Please check the parameter dimensions, %d parameters are too large for this GPU',
					e, self.params.size)
				exit(1)
		return gpu_data#}}}

	def release_gpumem(self, gpu_data):
		for name, array in gpu_data.items():
			try:
				gpu_data[name].gpudata.free()
			except drv.MemoryError as e:
				self.logger.error('%s.\n\t Freeing mem error', e)
				exit(1)

	def gpu_mem_info(self):

		cmd = "nvidia-smi -q -d MEMORY"#,UTILIZATION"
		os.system(cmd)  # returns the exit code in unix

	def run_simulation(self):

		# setup data#{{{
		data = { 'weights': self.weights, 'lengths': self.lengths, 'params': self.params.T }
		base_shape = self.n_work_items,
		for name, shape in dict(
			tavg0=(self.exposures, self.args.n_regions,),
			tavg1=(self.exposures, self.args.n_regions,),
			state=(self.buf_len, self.states * self.args.n_regions),
			).items():
			# memory error exception for compute device
			try:
				data[name] = np.zeros(shape + base_shape, 'f')
			except MemoryError as e:
				self.logger.error('%s.\n\t Please check the parameter dimensions %d x %d, they are to large '
							 'for this compute device',
							 e, self.args.n_sweep_arg0, self.args.n_sweep_arg1)
				exit(1)

		gpu_data = self.make_gpu_data(data)#{{{

		# setup CUDA stuff#{{{
		step_fn = self.make_kernel(
			source_file=self.args.filename,
			warp_size=32,
			# block_dim_x=self.args.n_sweep_arg0,
			# ext_options=preproccesor_defines,
			# caching=args.caching,
			args=self.args,
			lineinfo=self.args.lineinfo,
			nh=self.buf_len,
			)#}}}

		# setup simulation#{{{
		tic = time.time()

		n_streams = 32
		streams = [drv.Stream() for i in range(n_streams)]
		events = [drv.Event() for i in range(n_streams)]
		tavg_unpinned = []

		try:
			tavg = drv.pagelocked_zeros((n_streams,) + data['tavg0'].shape, dtype=np.float32)
		except drv.MemoryError as e:
			self.logger.error(
				'%s.\n\t Please check the parameter dimensions, %d parameters are too large for this GPU',
				e, self.params.size)
			exit(1)

		# determine optimal grid recursively
		def dog(fgd):
			maxgd, mingd = max(fgd), min(fgd)
			maxpos = fgd.index(max(fgd))
			if (maxgd - 1) * mingd * bx * by >= nwi:
				fgd[maxpos] = fgd[maxpos] - 1
				dog(fgd)
			else:
				return fgd

		# n_sweep_arg0 scales griddim.x, n_sweep_arg1 scales griddim.y
		# form an optimal grid recursively
		bx, by = self.args.blockszx, self.args.blockszy
		nwi = self.n_work_items
		rootnwi = int(np.ceil(np.sqrt(nwi)))
		gridx = int(np.ceil(rootnwi / bx))
		gridy = int(np.ceil(rootnwi / by))

		final_block_dim = bx, by, 1

		fgd = [gridx, gridy]
		dog(fgd)
		final_grid_dim = fgd[0], fgd[1]

		assert gridx * gridy * bx * by >= nwi

		self.logger.info('history shape %r', gpu_data['state'].shape)
		self.logger.info('gpu_data %s', gpu_data['tavg0'].shape)
		self.logger.info('on device mem: %.3f MiB' % (self.nbytes(data) / 1024 / 1024, ))
		self.logger.info('final block dim %r', final_block_dim)
		self.logger.info('final grid dim %r', final_grid_dim)

		# run simulation#{{{
		nstep = self.args.n_time

		self.gpu_mem_info() if self.args.verbose else None

		try:
			for i in tqdm.trange(nstep, file=sys.stdout):

				try:
					event = events[i % n_streams]
					stream = streams[i % n_streams]

					if i > 0:
						stream.wait_for_event(events[(i - 1) % n_streams])

					step_fn(np.uintc(i * self.n_inner_steps), np.uintc(self.args.n_regions), np.uintc(self.buf_len),
							np.uintc(self.n_inner_steps), np.uintc(self.n_work_items), np.float32(self.dt),
							gpu_data['weights'], gpu_data['lengths'], gpu_data['params'], gpu_data['state'],
							gpu_data['tavg%d' % (i%2,)],
							block=final_block_dim, grid=final_grid_dim)

					event.record(streams[i % n_streams])
				except drv.LaunchError as e:
					self.logger.error('%s', e)
					exit(1)

				tavgk = 'tavg%d' % ((i + 1) % 2,)

				# async wrt. other streams & host, but not this stream.
				if i >= n_streams:
					stream.synchronize()
					tavg_unpinned.append(tavg[i % n_streams].copy())

				drv.memcpy_dtoh_async(tavg[i % n_streams], gpu_data[tavgk].ptr, stream=stream)

			# recover uncopied data from pinned buffer
			if nstep > n_streams:
				for i in range(nstep % n_streams, n_streams):
					stream.synchronize()
					tavg_unpinned.append(tavg[i].copy())

			for i in range(nstep % n_streams):
				stream.synchronize()
				tavg_unpinned.append(tavg[i].copy())

		except drv.LogicError as e:
			self.logger.error('%s. Check the number of states of the model or '
						 'GPU block shape settings blockdim.x/y %r, griddim %r.',
						 e, final_block_dim, final_grid_dim)
			exit(1)
		except drv.RuntimeError as e:
			self.logger.error('%s', e)
			exit(1)


		# self.logger.info('kernel finish..')
		# release pinned memory
		tavg = np.array(tavg_unpinned)

		# also release gpu_data
		self.release_gpumem(gpu_data)

		self.logger.info('kernel finished')
		return tavg

	def plot_output(self, tavg):
		plt.plot((tavg[:, self.args.plot_data, :, 0]), 'k', alpha=.2)
		plt.show()

	def write_output(self, tavg):
		tavg_file = open('tavg_data', 'wb')
		pickle.dump(tavg, tavg_file)
		tavg_file.close()

	def run_all(self):

		np.random.seed(79)

		tic = time.time()

		tavg0 = self.run_simulation()
		toc = time.time()
		elapsed = toc - tic

		if (self.args.validate == True):
			self.compare_with_ref(tavg0)

		self.plot_output(tavg0) if self.args.plot_data is not None else None
		self.write_output(tavg0) if self.args.write_data else None
		self.logger.info('Output shape (simsteps, states, bnodes, n_params) %s', tavg0.shape)
		self.logger.info('Finished CUDA simulation successfully in: {0:.3f}'.format(elapsed))
		self.logger.info('and in {0:.3f} M step/s'.format(
			1e-6 * self.args.n_time * self.n_inner_steps * self.n_work_items / elapsed))

		return tavg0


if __name__ == '__main__':

	driver_setup = Driver_Setup()
	tavgGPU = Driver_Execute(driver_setup).run_all()

	simtime = driver_setup.args.n_time
	# simtime = 10
	regions = driver_setup.args.n_regions
	g = 1.0
	# g = 0.0042
	s = 1.0
	dt = driver_setup.dt
	period = 1

	# generic model definition
	model = driver_setup.args.model.capitalize()+'T'

	# non generic model names
	# model = 'MontbrioT'
	# model = 'RwongwangT'
	# model = 'OscillatorT'
	# model = 'DumontGutkin'
	# model = 'MontbrioPazoRoxin'
	# model='Generic2dOscillator'
	(time, tavgCPU) = regularRun(simtime, g, s, dt, period).simulate_python(model)

	print('CPUshape', tavgCPU.shape)
	print('GPUshape', tavgGPU.shape)

	# check for deviation tolerance between GPU and CPU
	# for basic coupling and period = 1
	# using euler deterministic solver
	max_err = []
	x = 0
	for t in range(0, simtime):
		# print(t, 'tol:', np.max(np.abs(actual[t] - expected[t, :, :, 0])))
		# print(t, 'tol:', np.max(np.abs(tavgCPU[t,:,:,0], tavgGPU[t,:,:,0])))
		print(t)
		# print('C', tavgCPU[t,:,:,0])
		# print('G', tavgGPU[t,:,:,0])
		# print(t, 'tol:', np.max(np.abs(tavgCPU[t,:,:,0] - tavgGPU[t,:,:,0])))
		np.testing.assert_allclose(tavgCPU[t, :, :, 0], tavgGPU[t, :, :, 0], 2e-5 * t * 2, 1e-5 * t * 2)


from tvb.simulator.models.base import Model, ModelNumbaDfun
import numpy
from numpy import *
from numba import guvectorize, float64
from tvb.basic.neotraits.api import NArray, Final, List, Range

class RwongwangT(ModelNumbaDfun):
        
    a_E = NArray(
        label=":math:`a_E`",
        default=numpy.array([310.0]),
        domain=Range(lo=0., hi=500., step=1.),
        doc=""""""
    )    
        
    b_E = NArray(
        label=":math:`b_E`",
        default=numpy.array([125.0]),
        domain=Range(lo=0., hi=200., step=1.),
        doc=""""""
    )    
        
    d_E = NArray(
        label=":math:`d_E`",
        default=numpy.array([0.160]),
        domain=Range(lo=0.0, hi=0.2, step=0.001),
        doc=""""""
    )    
        
    a_I = NArray(
        label=":math:`a_I`",
        default=numpy.array([615.0]),
        domain=Range(lo=0., hi=1000., step=1.),
        doc=""""""
    )    
        
    b_I = NArray(
        label=":math:`b_I`",
        default=numpy.array([177.0]),
        domain=Range(lo=0.0, hi=200.0, step=1.0),
        doc=""""""
    )    
        
    d_I = NArray(
        label=":math:`d_I`",
        default=numpy.array([0.087]),
        domain=Range(lo=0.01, hi=0.2, step=0.001),
        doc=""""""
    )    
        
    gamma_E = NArray(
        label=":math:`gamma_E`",
        default=numpy.array([0.641 / 1000.0]),
        domain=Range(lo=0.0, hi=0.001, step=0.00001),
        doc=""""""
    )    
        
    tau_E = NArray(
        label=":math:`tau_E`",
        default=numpy.array([100.0]),
        domain=Range(lo=50., hi=150., step=1.),
        doc=""""""
    )    
        
    w_plus = NArray(
        label=":math:`w_plus`",
        default=numpy.array([1.4]),
        domain=Range(lo=0.0, hi=2.0, step=0.01),
        doc=""""""
    )    
        
    w_E = NArray(
        label=":math:`w_E`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.0, hi=2.0, step=0.01),
        doc=""""""
    )    
        
    gamma_I = NArray(
        label=":math:`gamma_I`",
        default=numpy.array([1.0 / 1000.0]),
        domain=Range(lo=0.0, hi=0.002, step=0.0001),
        doc=""""""
    )    
        
    tau_I = NArray(
        label=":math:`tau_I`",
        default=numpy.array([10.0]),
        domain=Range(lo=5., hi=150., step=1.0),
        doc=""""""
    )    
        
    w_I = NArray(
        label=":math:`w_I`",
        default=numpy.array([0.7]),
        domain=Range(lo=0.0, hi=1.0, step=0.01),
        doc=""""""
    )    
        
    I_0 = NArray(
        label=":math:`I_0`",
        default=numpy.array([0.382]),
        domain=Range(lo=0.0, hi=1.0, step=0.001),
        doc=""""""
    )    
        
    J_N = NArray(
        label=":math:`J_N`",
        default=numpy.array([0.15]),
        doc=""""""
    )    
        
    J_I = NArray(
        label=":math:`J_I`",
        default=numpy.array([1.0]),
        doc=""""""
    )    
        
    G = NArray(
        label=":math:`G`",
        default=numpy.array([2.0]),
        doc=""""""
    )    
        
    lamda = NArray(
        label=":math:`lamda`",
        default=numpy.array([0.0]),
        doc=""""""
    )    
        
    J_NMDA = NArray(
        label=":math:`J_NMDA`",
        default=numpy.array([0.15]),
        doc=""""""
    )    
        
    JI = NArray(
        label=":math:`JI`",
        default=numpy.array([1.0]),
        doc=""""""
    )    

    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={"V": numpy.array([0.0, 1.0]), 
				 "W": numpy.array([0.0, 1.0])},
        doc="""state variables"""
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={"V": numpy.array([0.0, 1.0]), 
				 "W": numpy.array([0.0, 1.0])},
    )
    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=('V', 'W', ),
        default=('V', 'W', ),

        doc="Variables to monitor"
    )

    state_variables = ['V', 'W']

    _nvar = 2
    cvar = numpy.array([0,1,], dtype = numpy.int32)

    def dfun(self, vw, c, local_coupling=0.0):
        vw_ = vw.reshape(vw.shape[:-1]).T
        c_ = c.reshape(c.shape[:-1]).T
        deriv = _numba_dfun_RwongwangT(vw_, c_, self.a_E, self.b_E, self.d_E, self.a_I, self.b_I, self.d_I, self.gamma_E, self.tau_E, self.w_plus, self.w_E, self.gamma_I, self.tau_I, self.w_I, self.I_0, self.J_N, self.J_I, self.G, self.lamda, self.J_NMDA, self.JI, local_coupling)

        return deriv.T[..., numpy.newaxis]

@guvectorize([(float64[:], float64[:], float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64[:])], '(n),(m)' + ',()'*21 + '->(n)', nopython=True)
def _numba_dfun_RwongwangT(vw, coupling, a_E, b_E, d_E, a_I, b_I, d_I, gamma_E, tau_E, w_plus, w_E, gamma_I, tau_I, w_I, I_0, J_N, J_I, G, lamda, J_NMDA, JI, local_coupling, dx):
    "Gufunc for RwongwangT model equations."

    # long-range coupling
    c_pop0 = coupling[0]
    c_pop1 = coupling[1]

    V = vw[0]
    W = vw[1]

    # derived variables
    tmp_I_E = a_E * ((w_E * I_0) + (w_plus * J_NMDA) * V + c_pop0 - JI*W) - b_E
    tmp_H_E = tmp_I_E/(1.0-exp((-1.0 * d_E) * tmp_I_E))
    tmp_I_I = (a_I*(((w_I * I_0)+(J_NMDA * V))-W))-b_I
    tmp_H_I = tmp_I_I/(1.0-exp((-1.0 * d_I)*tmp_I_I))

    dx[0] = ((-1.0 / tau_E)* V)+(tmp_H_E*(1-V)*gamma_E)
    dx[1] = ((-1.0 / tau_I)* W)+(tmp_H_I*gamma_I)
    



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
module docstring
.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""

from tvb.datatypes.region_mapping import RegionVolumeMapping
from tvb.basic.neotraits.api import HasTraits, Attr, NArray

TRACTS_CHUNK_SIZE = 100


class Tracts(HasTraits):
    """Datatype for results of diffusion imaging tractography."""

    vertices = NArray(
        label="Vertex positions",
        doc="""An array specifying coordinates for the tracts vertices."""
    )

    tract_start_idx = NArray(
        dtype=int,
        label="Tract starting indices",
        doc="""Where is the first vertex of a tract in the vertex array"""
    )

    tract_region = NArray(
        dtype=int,
        label="Tract region index",
        required=False,
        doc="""
            An index used to find quickly all tract emerging from a region
            tract_region[i] is the region of the i'th tract. -1 represents the background
            """
    )

    region_volume_map = Attr(
        field_type=RegionVolumeMapping,
        label="Region volume Mapping used to create the tract_region index",
    )

    @property
    def tracts_count(self):
        return len(self.tract_start_idx) - 1



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
Created on Mar 20, 2013

.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

import numpy
from tvb.tests.library.base_testcase import BaseTestCase
from tvb.datatypes import temporal_correlations, time_series


class TestTemporalCorrelations(BaseTestCase):
    """
    Tests the defaults for `tvb.datatypes.temporal_correlations` module.
    """

    def test_crosscorrelation(self):
        data = numpy.random.random((10, 10))
        ts = time_series.TimeSeries(data=data, title='meh')
        dt = temporal_correlations.CrossCorrelation(source=ts, array_data=numpy.array([0]))
        summary_info = dt.summary_info()
        assert summary_info['Dimensions'] == ('Offsets', 'Node', 'Node', 'State Variable', 'Mode')
        assert summary_info['Source'] == 'meh'
        assert summary_info['Temporal correlation type'] == 'CrossCorrelation'
        assert dt.labels_ordering == ('Offsets', 'Node', 'Node', 'State Variable', 'Mode')
        assert dt.source is not None
        assert dt.time is None



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
LEMS2python module implements a DSL code generation using a TVB-specific LEMS-based DSL.

.. moduleauthor:: Michiel. A. van der Vlag <m.van.der.vlag@fz-juelich.de>   
.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>


        Usage: - create an modelfile.xml
               - import rateML
               - make instance: rateml(model_filename, language, your/XMLfolder, 'your/generatedModelsfolder')
        the current supported model framework languages are python and cuda.
        for new models models/__init.py__ is auto_updated if model is unfamiliar to tvb for py models
        model_class_name is the model_filename + 'T', so not to overwrite existing models. Be sure to add the t
        when simulating the model in TVB
        example model files:
            epileptor.xml
            generic2doscillatort.xml
            kuramoto.xml
            montbrio.xml
            reducedwongwang.xml

"""

import os
import re
import argparse
import numpy as np
import tvb.simulator.models
from mako.template import Template
from tvb.basic.logger.builder import get_logger
from lems.model.model import Model

logger = get_logger(__name__)


class RateML:

    def __init__(self, model_filename=None, language=None, XMLfolder=None, GENfolder=None):

        self.args = self.parse_args()

        self.model_filename = model_filename or self.args.model

        language = language or self.args.language.lower()

        try:
            assert language in ('cuda', 'python', 'Cuda', 'Python', 'CUDA')
            self.language = language.lower()
        except AssertionError as e:
            logger.error('Please choose between Python or Cuda %s', e)
            exit()

        self.XMLfolder = XMLfolder or self.args.source
        self.GENfolder = GENfolder or self.args.destination

        # set file locations
        self.generated_model_location = self.set_generated_model_location()
        self.xml_location = self.set_XML_model_folder()

        # start templating
        model_str, driver_str = self.render()

        # write model to user submitted location
        self.write_model_file(self.generated_model_location, model_str)

        if self.language == 'cuda':
            # write driver file to fixed ./run/ location
            self.write_model_file(self.set_driver_location(), driver_str)
            # for driver robustness also write XML to default location generatedModels folder for CUDA
            if GENfolder != None:
                default_save = os.path.join(self.default_generation_folder(), self.model_filename.lower() + '.c')
                self.write_model_file(default_save, model_str)

        # if it is a TVB.py model, it should be familiarized
        if self.language.lower() == 'python':
            self.familiarize_TVB(model_str)

    def parse_args(self):  # {{{
        parser = argparse.ArgumentParser(description='Run XML model conversion.')

        parser.add_argument('-m', '--model', default='rwongwang',
                            help="neural mass model to be converted")
        parser.add_argument('-l', '--language', default='Python',
                            help="programming language output")
        parser.add_argument('-s', '--source', default='',
                            help="source folder location")
        parser.add_argument('-d', '--destination', default='',
                            help="destination folder location")

        args = parser.parse_args()
        return args

    @staticmethod
    def default_XML_folder():
        here = os.path.dirname(os.path.abspath(__file__))
        xmlpath = os.path.join(here, 'XMLmodels')
        return xmlpath

    def set_XML_model_folder(self):
        folder = self.XMLfolder or self.default_XML_folder()

        try:
            location = os.path.join(folder, self.model_filename.lower() + '.xml')
            assert os.path.isfile(location)
        except AssertionError:
            logger.error('XML folder %s does not contain %s', folder, self.model_filename + '.xml')
            exit()

        return location

    @staticmethod
    def default_generation_folder():
        here = os.path.dirname(os.path.abspath(__file__))
        modelpath = os.path.join(here, 'generatedModels')
        return modelpath

    def set_generated_model_location(self):
        folder = self.GENfolder or self.default_generation_folder()
        lan = self.language.lower()
        if lan == 'python':
            ext = '.py'
        elif lan == 'cuda':
            ext = '.c'

        try:
            location = os.path.join(folder)
            assert os.path.isdir(location)
        except AssertionError:
            logger.error('Generation folder %s does not exist', location)
            exit()

        return os.path.join(folder, self.model_filename.lower() + ext)

    def set_driver_location(self):
        here = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(here, 'run', 'model_driver_' + self.model_filename + '.py')

    def set_template(self, name):
        here = os.path.dirname(os.path.abspath(__file__))
        tmp_filename = os.path.join(here, 'tmpl8_' + name + '.py')
        template = Template(filename=tmp_filename)
        return template

    def XSD_validate_XML(self):

        ''' Use own validation instead of LEMS because of slight difference in definition file'''

        from lxml import etree
        from urllib.request import urlopen

        # Global XSD file location
        schema_file = urlopen(
            "https://raw.githubusercontent.com/the-virtual-brain/tvb-root/master/scientific_library/tvb/rateML/rML_v0.xsd")
        xmlschema = etree.XMLSchema(etree.parse(schema_file))
        xmlschema.assertValid(etree.parse(self.xml_location))
        logger.info("True validation of {0} against {1}".format(self.xml_location, schema_file.geturl()))

    def pp_bound(self, model):

        # check if boundaries for state variables are present. contruct is not necessary in pymodels
        # python only
        svboundaries = False
        for i, sv in enumerate(model.component_types['derivatives'].dynamics.state_variables):
            if sv.exposure != 'None' and sv.exposure != '' and sv.exposure:
                svboundaries = True
                continue

        return svboundaries

    def pp_cplist(self, model):

        # check for component_types containing coupling in name and gather data.
        # multiple coupling functions could be defined in xml
        # cuda only
        couplinglist = list()
        for i, cplists in enumerate(model.component_types):
            if 'coupling' in cplists.name:
                couplinglist.append(cplists)

        return couplinglist

    def pp_noise(self, model):

        # only check whether noise is there, if so then activate it
        # cuda only
        noisepresent = False
        for ct in (model.component_types):
            if ct.name == 'noise':
                noisepresent = True

        # see if nsig derived parameter is present for noise
        # cuda only
        modellist = model.component_types['derivatives']
        nsigpresent = False
        if noisepresent == True:
            for dprm in (modellist.derived_parameters):
                if (dprm.name == 'nsig' or dprm.name == 'NSIG'):
                    nsigpresent = True

        return noisepresent, nsigpresent

    # check for power symbol and parse to python (**) or c power (powf(x, y))
    def swap_language_specific_terms(self, model_str):

        if self.language == 'cuda':
            model_str = re.sub(r"\bpi\b", 'PI', model_str)
            model_str = re.sub(r"\binf\b", 'INF', model_str)

        for power in re.finditer(r"\{(.*?)(\^)(.*?)\}", model_str):
            target = power.group(0)
            powersplit = target.split('^')
            if self.language == 'cuda':
                pow = 'powf(' + powersplit[0].replace('{', '') + ', ' + powersplit[1].replace('}', '') + ')'
            if self.language == 'python':
                pow = powersplit[0].replace('{', '') + '**' + powersplit[1].replace('}', '')
            model_str = re.sub(re.escape(target), pow, model_str)

        return model_str

    # setting the inital value for cuda models
    # the entered range is splitted and a random value is generated within range
    # if values are equal then that is the inital value
    def init_statevariables(self, model):

        modellist = model.component_types['derivatives'].dynamics.state_variables
        for sv in modellist:
            splitdim = list(sv.dimension.split(","))
            sv_rnd = np.random.uniform(low=float(splitdim[0]), high=float(splitdim[1]))
            model.component_types['derivatives'].dynamics.state_variables[sv.name].dimension = sv_rnd

    def load_model(self):
        "Load model from filename"

        # instantiate LEMS lib
        model = Model()
        model.import_from_file(self.xml_location)

        self.XSD_validate_XML()

        # Do some preprocessing on the template to easify rendering
        noisepresent, nsigpresent = self.pp_noise(model)
        couplinglist = self.pp_cplist(model)
        svboundaries = self.pp_bound(model)

        if self.language == 'cuda':
            self.init_statevariables(model)

        return model, svboundaries, couplinglist, noisepresent, nsigpresent

    def render(self):
        '''
        render_model start the mako templating.
        this function is similar for all languages. its .render arguments are overloaded.
        '''

        model, svboundaries, couplinglist, noisepresent, nsigpresent = self.load_model()

        derivative_list = model.component_types['derivatives']

        model_str = self.render_model(derivative_list, svboundaries, couplinglist, noisepresent, nsigpresent)

        model_str = self.swap_language_specific_terms(model_str)

        # render driver only in case of cuda
        if self.language == 'cuda':
            driver_str = self.render_driver(derivative_list)
        else:
            driver_str = None

        return model_str, driver_str

    def render_model(self, derivative_list, svboundaries, couplinglist, noisepresent, nsigpresent):

        if self.language == 'python':
            model_class_name = self.model_filename.capitalize() + 'T'
        if self.language == 'cuda':
            model_class_name = self.model_filename

        # start templating
        model_str = self.set_template(self.language).render(
            modelname=model_class_name,  # all
            const=derivative_list.constants,  # all
            dynamics=derivative_list.dynamics,  # all
            exposures=derivative_list.exposures,  # all
            params=derivative_list.parameters,  # cuda
            derparams=derivative_list.derived_parameters,  # cuda
            svboundaries=svboundaries,  # python
            coupling=couplinglist,  # cuda
            noisepresent=noisepresent,  # cuda
            nsigpresent=nsigpresent,  # cuda
        )

        return model_str

    def render_driver(self, derivative_list):

        driver_str = self.set_template('driver').render(
            model=self.model_filename,
            XML=derivative_list,
        )

        return driver_str

    def familiarize_TVB(self, model_str):
        '''
        Write new model to TVB model location and into init.py such it is familiar to TVB if not already present
        This is for Python models only
        '''

        model_filename = self.model_filename
        # set tvb location
        TVB_model_location = os.path.join(os.path.dirname(tvb.simulator.models.__file__),
                                          model_filename.lower() + 'T.py')
        # next to user submitted location also write to default tvb location
        self.write_model_file(TVB_model_location, model_str)

        try:
            doprint = True
            modelenumnum = 0
            modulemodnum = 0
            with open(os.path.join(os.path.dirname(tvb.simulator.models.__file__), '__init__.py'), "r+") as f:
                lines = f.readlines()
                for num, line in enumerate(lines):
                    if (model_filename.upper() + 'T = ' + "\"" + model_filename.capitalize() + "T\"") in line:
                        doprint = False
                    elif ("class ModelsEnum(Enum):") in line:
                        modelenumnum = num
                    elif ("_module_models = {") in line:
                        modulemodnum = num
                if doprint:
                    lines.insert(modelenumnum + 1, "    " + model_filename.upper() + 'T = ' + "\"" +
                                 model_filename.capitalize() + "T\"\n")
                    lines.insert(modulemodnum + 2, "    " + "'" + model_filename.lower() + "T'" + ': '
                                 + "[ModelsEnum." + model_filename.upper() + "T],\n")
                    f.truncate(0)
                    f.seek(0)
                    f.writelines(lines)
                logger.info("model file generated {}".format(model_filename))
        except IOError as e:
            logger.error('Writing TVB model to file failed: %s', e)

    def write_model_file(self, model_location, model_str):

        '''Write templated model to file'''

        try:
            with open(model_location, "w") as f:
                f.writelines(model_str)
        except IOError as e:
            logger.error('Writing %s model to file failed: %s', self.language, e)


if __name__ == "__main__":
    RateML()

    # example for direct project implementation
    # set the language for your model
    # language='python'
    # language='Cuda'

    # choose an example or your own model
    # model_filename = 'montbrio'
    # model_filename = 'oscillator'
    # model_filename = 'kuramoto'
    # model_filename = 'rwongwang'
    # model_filename = 'epileptor'

    # start conversion to default model location
    # RateML(model_filename, language)



from tvb.simulator.models.base import Model, ModelNumbaDfun
import numpy
from numpy import *
from numba import guvectorize, float64
from tvb.basic.neotraits.api import NArray, Final, List, Range

class EpileptorT(ModelNumbaDfun):
        
    a = NArray(
        label=":math:`a`",
        default=numpy.array([1.0]),
        doc=""""""
    )    
        
    b = NArray(
        label=":math:`b`",
        default=numpy.array([3.0]),
        doc=""""""
    )    
        
    c = NArray(
        label=":math:`c`",
        default=numpy.array([1.0]),
        doc=""""""
    )    
        
    d = NArray(
        label=":math:`d`",
        default=numpy.array([5.0]),
        doc=""""""
    )    
        
    r = NArray(
        label=":math:`r`",
        default=numpy.array([0.00035]),
        domain=Range(lo=0.0, hi=0.001, step=0.00005),
        doc=""""""
    )    
        
    s = NArray(
        label=":math:`s`",
        default=numpy.array([4.0]),
        doc=""""""
    )    
        
    x0 = NArray(
        label=":math:`x0`",
        default=numpy.array([-1.6]),
        domain=Range(lo=-3.0, hi=-1.0, step=0.1),
        doc=""""""
    )    
        
    Iext = NArray(
        label=":math:`Iext`",
        default=numpy.array([3.1]),
        domain=Range(lo=1.5, hi=5.0, step=0.1),
        doc=""""""
    )    
        
    slope = NArray(
        label=":math:`slope`",
        default=numpy.array([0.]),
        domain=Range(lo=-16.0, hi=6.0, step=0.1),
        doc=""""""
    )    
        
    Iext2 = NArray(
        label=":math:`Iext2`",
        default=numpy.array([0.45]),
        domain=Range(lo=0.0, hi=1.0, step=0.05),
        doc=""""""
    )    
        
    tau = NArray(
        label=":math:`tau`",
        default=numpy.array([10.0]),
        doc=""""""
    )    
        
    aa = NArray(
        label=":math:`aa`",
        default=numpy.array([6.0]),
        doc=""""""
    )    
        
    bb = NArray(
        label=":math:`bb`",
        default=numpy.array([2.0]),
        doc=""""""
    )    
        
    Kvf = NArray(
        label=":math:`Kvf`",
        default=numpy.array([0.0]),
        domain=Range(lo=0.0, hi=4.0, step=0.5),
        doc=""""""
    )    
        
    Kf = NArray(
        label=":math:`Kf`",
        default=numpy.array([0.0]),
        domain=Range(lo=0.0, hi=4.0, step=0.5),
        doc=""""""
    )    
        
    Ks = NArray(
        label=":math:`Ks`",
        default=numpy.array([0.0]),
        domain=Range(lo=-4.0, hi=4.0, step=0.1),
        doc=""""""
    )    
        
    tt = NArray(
        label=":math:`tt`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.001, hi=10.0, step=0.001),
        doc=""""""
    )    
        
    modification = NArray(
        label=":math:`modification`",
        default=numpy.array([0]),
        doc=""""""
    )    

    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={"x1": numpy.array([0.0]), 
				 "y1": numpy.array([0.0]), 
				 "z": numpy.array([0.0]), 
				 "x2": numpy.array([0.0]), 
				 "y2": numpy.array([0.0]), 
				 "g": numpy.array([0.0])},
        doc="""state variables"""
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={"x1": numpy.array([-2.0, 1.0]), 
				 "y1": numpy.array([-20.0, 2.0]), 
				 "z": numpy.array([-2.0, 5.0]), 
				 "x2": numpy.array([-2.0, 0.0]), 
				 "y2": numpy.array([0.0, 2.0]), 
				 "g": numpy.array([-1.0, 1.0])},
    )
    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=('x1 ** x2', 'x2', ),
        default=('x1', 'y1', 'z', 'x2', 'y2', 'g', ),
        doc="Variables to monitor"
    )

    state_variables = ['x1', 'y1', 'z', 'x2', 'y2', 'g']

    _nvar = 6
    cvar = numpy.array([0,1,2,3,4,5,], dtype = numpy.int32)

    def dfun(self, vw, c, local_coupling=0.0):
        vw_ = vw.reshape(vw.shape[:-1]).T
        c_ = c.reshape(c.shape[:-1]).T
        deriv = _numba_dfun_EpileptorT(vw_, c_, self.a, self.b, self.c, self.d, self.r, self.s, self.x0, self.Iext, self.slope, self.Iext2, self.tau, self.aa, self.bb, self.Kvf, self.Kf, self.Ks, self.tt, self.modification, local_coupling)

        return deriv.T[..., numpy.newaxis]

@guvectorize([(float64[:], float64[:], float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64[:])], '(n),(m)' + ',()'*19 + '->(n)', nopython=True)
def _numba_dfun_EpileptorT(vw, coupling, a, b, c, d, r, s, x0, Iext, slope, Iext2, tau, aa, bb, Kvf, Kf, Ks, tt, modification, local_coupling, dx):
    "Gufunc for EpileptorT model equations."

    # long-range coupling
    c_pop0 = coupling[0]
    c_pop1 = coupling[1]
    c_pop2 = coupling[2]
    c_pop3 = coupling[3]
    c_pop4 = coupling[4]
    c_pop5 = coupling[5]

    x1 = vw[0]
    y1 = vw[1]
    z = vw[2]
    x2 = vw[3]
    y2 = vw[4]
    g = vw[5]


    # Conditional variables
    if x1 < 0.0:
        ydot0 = -a * x1 ** 2 + b * x1
    else:
        ydot0 = slope - x2 + 0.6 * z-4 ** 2

    if z < 0.0:
        ydot2 = - 0.1 * z ** 7
    else:
        ydot2 = 0

    if modification:
        h = x0 + 3. / (1. + exp(-(x1 + 0.5) / 0.1))
    else:
        h = 4 * (x1 - x0) + ydot2

    if x2 < -0.25:
        ydot4 = 0.0
    else:
        ydot4 = aa * (x2 + 0.25)

    dx[0] = tt * (y1 - z + Iext + Kvf * c_pop1 + ydot0 )
    dx[1] = tt * (c - d * x1 ** 2 - y1)
    dx[2] = tt * (r * (h - z + Ks * c_pop1))
    dx[3] = tt * (-y2 + x2 - x2 ** 3 + Iext2 + bb * g - 0.3 * (z - 3.5) + Kf * c_pop2)
    dx[4] = tt * (-y2 + ydot4) / tau
    dx[5] = tt * (-0.01 * (g - 0.1 * x1) )
    



from tvb.simulator.models.base import Model, ModelNumbaDfun
import numpy
from numpy import *
from numba import guvectorize, float64
from tvb.basic.neotraits.api import NArray, Final, List, Range

class OscillatorT(ModelNumbaDfun):
        
    tau = NArray(
        label=":math:`tau`",
        default=numpy.array([1.0]),
        doc=""""""
    )    
        
    I = NArray(
        label=":math:`I`",
        default=numpy.array([0.1]),
        doc=""""""
    )    
        
    a = NArray(
        label=":math:`a`",
        default=numpy.array([0.5]),
        doc=""""""
    )    
        
    b = NArray(
        label=":math:`b`",
        default=numpy.array([0.4]),
        doc=""""""
    )    
        
    c = NArray(
        label=":math:`c`",
        default=numpy.array([-4.0]),
        doc=""""""
    )    
        
    d = NArray(
        label=":math:`d`",
        default=numpy.array([0.02]),
        doc=""""""
    )    
        
    e = NArray(
        label=":math:`e`",
        default=numpy.array([3.0]),
        doc=""""""
    )    
        
    f = NArray(
        label=":math:`f`",
        default=numpy.array([1.0]),
        doc=""""""
    )    
        
    g = NArray(
        label=":math:`g`",
        default=numpy.array([0.0]),
        doc=""""""
    )    
        
    alpha = NArray(
        label=":math:`alpha`",
        default=numpy.array([1.0]),
        doc=""""""
    )    
        
    beta = NArray(
        label=":math:`beta`",
        default=numpy.array([1.0]),
        doc=""""""
    )    
        
    gamma = NArray(
        label=":math:`gamma`",
        default=numpy.array([1.0]),
        doc=""""""
    )    

    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={"V": numpy.array([0.0, 0.0]), 
				 "W": numpy.array([0.0, 0.0])},
        doc="""state variables"""
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={"V": numpy.array([-2.0, 4.0]), 
				 "W": numpy.array([-6.0, 6.0])},
    )
    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=('V', 'W', ),
        default=('V', 'W', ),

        doc="Variables to monitor"
    )

    state_variables = ['V', 'W']

    _nvar = 2
    cvar = numpy.array([0,1,], dtype = numpy.int32)

    def dfun(self, vw, c, local_coupling=0.0):
        vw_ = vw.reshape(vw.shape[:-1]).T
        c_ = c.reshape(c.shape[:-1]).T
        deriv = _numba_dfun_OscillatorT(vw_, c_, self.tau, self.I, self.a, self.b, self.c, self.d, self.e, self.f, self.g, self.alpha, self.beta, self.gamma, local_coupling)

        return deriv.T[..., numpy.newaxis]

@guvectorize([(float64[:], float64[:], float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, float64[:])], '(n),(m)' + ',()'*13 + '->(n)', nopython=True)
def _numba_dfun_OscillatorT(vw, coupling, tau, I, a, b, c, d, e, f, g, alpha, beta, gamma, local_coupling, dx):
    "Gufunc for OscillatorT model equations."

    # long-range coupling
    c_pop0 = coupling[0]
    c_pop1 = coupling[1]

    V = vw[0]
    W = vw[1]


    dx[0] = d * tau * (alpha * W - f * V**3 + e * V**2 + g * V + gamma * I + gamma * c_pop0 * V)
    dx[1] = d * (a + b * V + c * V**2 - beta * W) / tau
    



from __future__ import print_function

import logging
import itertools
import argparse
import pickle

from tvb.simulator.lab import *
from tvb.basic.logger.builder import get_logger

from tvb.rateML.run.regular_run import regularRun

import os.path
import numpy as np
try:
	import pycuda.autoinit
	import pycuda.driver as drv
	from pycuda.compiler import SourceModule
	import pycuda.gpuarray as gpuarray
except ImportError:
	logging.warning('pycuda not available, rateML driver not usable.')

import matplotlib.pyplot as plt

import time
import tqdm

here = os.path.dirname(os.path.abspath(__file__))

class Driver_Setup:

	def __init__(self):
		self.args = self.parse_args()

		self.logger = get_logger('tvb.rateML')
		self.logger.setLevel(level='INFO' if self.args.verbose else 'WARNING')

		self.checkargbounds()

		self.dt = self.args.delta_time
		self.connectivity = self.tvb_connectivity(self.args.n_regions)
		self.weights = self.connectivity.weights
		self.lengths = self.connectivity.tract_lengths
		self.tavg_period = 1.0
		self.n_inner_steps = int(self.tavg_period / self.dt)

		self.params = self.setup_params(
		% for pc in range(len(XML.parameters)):
		self.args.n_sweep_arg${pc},
		% endfor
		)

		# bufferlength is based on the minimum of the first swept parameter (speed for many tvb models)
		self.n_work_items, self.n_params = self.params.shape
		self.buf_len_ = ((self.lengths / self.args.speeds_min / self.dt).astype('i').max() + 1)
		self.buf_len = 2 ** np.argwhere(2 ** np.r_[:30] > self.buf_len_)[0][0]  # use next power of

		self.states = self.args.states
		self.exposures = self.args.exposures

		if self.args.gpu_info:
			self.logger.setLevel(level='INFO')
			self.gpu_device_info()
			exit(1)

		self.logdata()

	def logdata(self):

		self.logger.info('dt %f', self.dt)
		self.logger.info('n_nodes %d', self.args.n_regions)
		self.logger.info('weights.shape %s', self.weights.shape)
		self.logger.info('lengths.shape %s', self.lengths.shape)
		self.logger.info('tavg period %s', self.tavg_period)
		self.logger.info('n_inner_steps %s', self.n_inner_steps)
		self.logger.info('params shape %s', self.params.shape)

		self.logger.info('nstep %d', self.args.n_time)
		self.logger.info('n_inner_steps %f', self.n_inner_steps)

		# self.logger.info('single connectome, %d x %d parameter space', self.args.n_sweep_arg0, self.args.n_sweep_arg1)
		self.logger.info('real buf_len %d, using power of 2 %d', self.buf_len_, self.buf_len)
		self.logger.info('number of states %d', self.states)
		self.logger.info('model %s', self.args.model)
		self.logger.info('real buf_len %d, using power of 2 %d', self.buf_len_, self.buf_len)
		self.logger.info('memory for states array on GPU %d MiB',
						 (self.buf_len * self.n_work_items * self.states * self.args.n_regions * 4) / 1024 ** 2)

	def checkargbounds(self):

		try:
			assert self.args.n_sweep_arg0 > 0, "Min value for [N_SWEEP_ARG0] is 1"
			assert self.args.n_time > 0, "Minimum number for [-n N_TIME] is 1"
			assert self.args.n_regions > 0, "Min value for  [-tvbn n_regions] for default data set is 68"
			assert self.args.blockszx > 0 and self.args.blockszx <= 32,	"Bounds for [-bx BLOCKSZX] are 0 < value <= 32"
			assert self.args.blockszy > 0 and self.args.blockszy <= 32, "Bounds for [-by BLOCKSZY] are 0 < value <= 32"
			assert self.args.delta_time > 0.0, "Min value for [-dt delta_time] is > 0.0, default is 0.1"
			assert self.args.speeds_min > 0.0, "Min value for [-sm speeds_min] is > 0.0, default is 3e-3"
			assert self.args.exposures > 0, "Min value for [-x exposures] is 1"
			assert self.args.states > 0, "Min value for [-s states] is 1"
		except AssertionError as e:
			self.logger.error('%s', e)
			raise

	def tvb_connectivity(self, tvbnodes):
		white_matter = connectivity.Connectivity.from_file(source_file="connectivity_"+str(tvbnodes)+".zip")
		# white_matter = connectivity.Connectivity.from_file(source_file="paupau.zip")
		white_matter.configure()
		return white_matter

	def parse_args(self):  # {{{
		parser = argparse.ArgumentParser(description='Run parameter sweep.')

		# for every parameter that needs to be swept, the size can be set
		% for pc in range(len(XML.parameters)):
		parser.add_argument('-s${pc}', '--n_sweep_arg${pc}', default=4, help='num grid points for ${pc+1}st parameter', type=int)
		% endfor
		parser.add_argument('-n', '--n_time', default=400, help='number of time steps to do', type=int)
		parser.add_argument('-v', '--verbose', default=False, help='increase logging verbosity', action='store_true')
		parser.add_argument('-m', '--model', default='${model}', help="neural mass model to be used during the simulation")
		parser.add_argument('-s', '--states', default=${XML.dynamics.state_variables.__len__()}, type=int, help="number of states for model")
		parser.add_argument('-x', '--exposures', default=${XML.exposures.__len__()}, type=int, help="number of exposures for model")
		parser.add_argument('-l', '--lineinfo', default=False, help='generate line-number information for device code.', action='store_true')
		parser.add_argument('-bx', '--blockszx', default=8, type=int, help="gpu block size x")
		parser.add_argument('-by', '--blockszy', default=8, type=int, help="gpu block size y")
		parser.add_argument('-val', '--validate', default=False, help="enable validation with refmodels", action='store_true')
		parser.add_argument('-r', '--n_regions', default="68", type=int, help="number of tvb nodes")
		parser.add_argument('-p', '--plot_data', type=int, help="plot res data for selected state")
		parser.add_argument('-w', '--write_data', default=False, help="write output data to file: 'tavg_data", action='store_true')
		parser.add_argument('-g', '--gpu_info', default=False, help="show gpu info", action='store_true')
		parser.add_argument('-dt', '--delta_time', default=0.1, type=float, help="dt for simulation")
		parser.add_argument('-sm', '--speeds_min', default=1, type=float, help="min speed for temporal buffer")

		args = parser.parse_args()
		return args


	def setup_params(self,
		% for pc in range(len(XML.parameters)):
		n${pc},
		% endfor
		):
		'''
		This code generates the parameters ranges that need to be set
		'''
		% for pc, par_var in enumerate(XML.parameters):
		sweeparam${pc} = np.linspace(${par_var.dimension}, n${pc})
		% endfor
		params = itertools.product(
		% for pc in range(len(XML.parameters)):
		sweeparam${pc},
		% endfor
		)
		params = np.array([vals for vals in params], np.float32)
		return params


	def gpu_device_info(self):
		'''
		Get GPU device information
		TODO use this information to give user GPU setting suggestions
		'''
		dev = drv.Device(0)
		print('\n')
		self.logger.info('GPU = %s', dev.name())
		self.logger.info('TOTAL AVAIL MEMORY: %d MiB', dev.total_memory()/1024/1024)

		# get device information
		att = {'MAX_THREADS_PER_BLOCK': [],
			   'MAX_BLOCK_DIM_X': [],
			   'MAX_BLOCK_DIM_Y': [],
			   'MAX_BLOCK_DIM_Z': [],
			   'MAX_GRID_DIM_X': [],
			   'MAX_GRID_DIM_Y': [],
			   'MAX_GRID_DIM_Z': [],
			   'TOTAL_CONSTANT_MEMORY': [],
			   'WARP_SIZE': [],
			   # 'MAX_PITCH': [],
			   'CLOCK_RATE': [],
			   'TEXTURE_ALIGNMENT': [],
			   # 'GPU_OVERLAP': [],
			   'MULTIPROCESSOR_COUNT': [],
			   'SHARED_MEMORY_PER_BLOCK': [],
			   'MAX_SHARED_MEMORY_PER_BLOCK': [],
			   'REGISTERS_PER_BLOCK': [],
			   'MAX_REGISTERS_PER_BLOCK': []}

		for key in att:
			getstring = 'drv.device_attribute.' + key
			# att[key].append(eval(getstring))
			self.logger.info(key + ': %s', dev.get_attribute(eval(getstring)))

class Driver_Execute(Driver_Setup):

	def __init__(self, ds):
		self.args = ds.args
		self.set_CUDAmodel_dir()
		self.weights, self.lengths, self.params = ds.weights, ds.lengths, ds.params
		self.buf_len, self.states, self.n_work_items = ds.buf_len, ds.states, ds.n_work_items
		self.n_inner_steps, self.n_params, self.dt = ds.n_inner_steps, ds.n_params, ds.dt
		self.exposures, self.logger = ds.exposures, ds.logger

	def set_CUDAmodel_dir(self):
		self.args.filename = os.path.join((os.path.dirname(os.path.abspath(__file__))),
								 "../generatedModels", self.args.model.lower() + '.c')

	def set_CUDA_ref_model_dir(self):
		self.args.filename = os.path.join((os.path.dirname(os.path.abspath(__file__))),
								 "../generatedModels/cuda_refs", self.args.model.lower() + '.c')

	def compare_with_ref(self, tavg0):
		self.args.model = self.args.model + 'ref'
		self.set_CUDA_ref_model_dir()
		tavg1 = self.run_simulation()

		# compare output to check if same as template
		# comparison = (tavg0.ravel() == tavg1.ravel())
		# self.logger.info('Templated version is similar to original %d:', comparison.all())
		self.logger.info('Corr.coef. of model with %s is: %f', self.args.model,
						 np.corrcoef(tavg0.ravel(), tavg1.ravel())[0, 1])

		return np.corrcoef(tavg0.ravel(), tavg1.ravel())[0, 1]

	def make_kernel(self, source_file, warp_size, args, lineinfo=False, nh='nh'):

		try:
			with open(source_file, 'r') as fd:
				source = fd.read()
				source = source.replace('M_PI_F', '%ff' % (np.pi, ))
				opts = ['--ptxas-options=-v', '-maxrregcount=32']
				if lineinfo:
					opts.append('-lineinfo')
				opts.append('-DWARP_SIZE=%d' % (warp_size, ))
				opts.append('-DNH=%s' % (nh, ))

				idirs = [here]
				self.logger.info('nvcc options %r', opts)

				try:
					network_module = SourceModule(
							source, options=opts, include_dirs=idirs,
							no_extern_c=True,
							keep=False,)
				except drv.CompileError as e:
					self.logger.error('Compilation failure \n %s', e)
					exit(1)

				# generic func signature creation
				mod_func = "{}{}{}{}".format('_Z', len(args.model), args.model, 'jjjjjfPfS_S_S_S_')

				step_fn = network_module.get_function(mod_func)

		except FileNotFoundError as e:
			self.logger.error('%s.\n  Generated model filename should match model on cmdline', e)
			exit(1)

		return step_fn #}}}

	def cf(self, array):#{{{
		# coerce possibly mixed-stride, double precision array to C-order single precision
		return array.astype(dtype='f', order='C', copy=True)#}}}

	def nbytes(self, data):#{{{
		# count total bytes used in all data arrays
		nbytes = 0
		for name, array in data.items():
			nbytes += array.nbytes
		return nbytes#}}}

	def make_gpu_data(self, data):#{{{
		# put data onto gpu
		gpu_data = {}
		for name, array in data.items():
			try:
				gpu_data[name] = gpuarray.to_gpu(self.cf(array))
			except drv.MemoryError as e:
				self.gpu_mem_info()
				self.logger.error(
					'%s.\n\t Please check the parameter dimensions, %d parameters are too large for this GPU',
					e, self.params.size)
				exit(1)
		return gpu_data#}}}

	def release_gpumem(self, gpu_data):
		for name, array in gpu_data.items():
			try:
				gpu_data[name].gpudata.free()
			except drv.MemoryError as e:
				self.logger.error('%s.\n\t Freeing mem error', e)
				exit(1)

	def gpu_mem_info(self):

		cmd = "nvidia-smi -q -d MEMORY"#,UTILIZATION"
		os.system(cmd)  # returns the exit code in unix

	def run_simulation(self):

		# setup data#{{{
		data = { 'weights': self.weights, 'lengths': self.lengths, 'params': self.params.T }
		base_shape = self.n_work_items,
		for name, shape in dict(
			tavg0=(self.exposures, self.args.n_regions,),
			tavg1=(self.exposures, self.args.n_regions,),
			state=(self.buf_len, self.states * self.args.n_regions),
			).items():
			# memory error exception for compute device
			try:
				data[name] = np.zeros(shape + base_shape, 'f')
			except MemoryError as e:
				self.logger.error('%s.\n\t Please check the parameter dimensions %d x %d, they are to large '
							 'for this compute device',
							 e, self.args.n_sweep_arg0, self.args.n_sweep_arg1)
				exit(1)

		gpu_data = self.make_gpu_data(data)#{{{

		# setup CUDA stuff#{{{
		step_fn = self.make_kernel(
			source_file=self.args.filename,
			warp_size=32,
			# block_dim_x=self.args.n_sweep_arg0,
			# ext_options=preproccesor_defines,
			# caching=args.caching,
			args=self.args,
			lineinfo=self.args.lineinfo,
			nh=self.buf_len,
			)#}}}

		# setup simulation#{{{
		tic = time.time()

		n_streams = 32
		streams = [drv.Stream() for i in range(n_streams)]
		events = [drv.Event() for i in range(n_streams)]
		tavg_unpinned = []

		try:
			tavg = drv.pagelocked_zeros((n_streams,) + data['tavg0'].shape, dtype=np.float32)
		except drv.MemoryError as e:
			self.logger.error(
				'%s.\n\t Please check the parameter dimensions, %d parameters are too large for this GPU',
				e, self.params.size)
			exit(1)

		# determine optimal grid recursively
		def dog(fgd):
			maxgd, mingd = max(fgd), min(fgd)
			maxpos = fgd.index(max(fgd))
			if (maxgd - 1) * mingd * bx * by >= nwi:
				fgd[maxpos] = fgd[maxpos] - 1
				dog(fgd)
			else:
				return fgd

		# n_sweep_arg0 scales griddim.x, n_sweep_arg1 scales griddim.y
		# form an optimal grid recursively
		bx, by = self.args.blockszx, self.args.blockszy
		nwi = self.n_work_items
		rootnwi = int(np.ceil(np.sqrt(nwi)))
		gridx = int(np.ceil(rootnwi / bx))
		gridy = int(np.ceil(rootnwi / by))

		final_block_dim = bx, by, 1

		fgd = [gridx, gridy]
		dog(fgd)
		final_grid_dim = fgd[0], fgd[1]

		assert gridx * gridy * bx * by >= nwi

		self.logger.info('history shape %r', gpu_data['state'].shape)
		self.logger.info('gpu_data %s', gpu_data['tavg0'].shape)
		self.logger.info('on device mem: %.3f MiB' % (self.nbytes(data) / 1024 / 1024, ))
		self.logger.info('final block dim %r', final_block_dim)
		self.logger.info('final grid dim %r', final_grid_dim)

		# run simulation#{{{
		nstep = self.args.n_time

		self.gpu_mem_info() if self.args.verbose else None

		try:
			for i in tqdm.trange(nstep, file=sys.stdout):

				try:
					event = events[i % n_streams]
					stream = streams[i % n_streams]

					if i > 0:
						stream.wait_for_event(events[(i - 1) % n_streams])

					step_fn(np.uintc(i * self.n_inner_steps), np.uintc(self.args.n_regions), np.uintc(self.buf_len),
							np.uintc(self.n_inner_steps), np.uintc(self.n_work_items), np.float32(self.dt),
							gpu_data['weights'], gpu_data['lengths'], gpu_data['params'], gpu_data['state'],
							gpu_data['tavg%d' % (i%2,)],
							block=final_block_dim, grid=final_grid_dim)

					event.record(streams[i % n_streams])
				except drv.LaunchError as e:
					self.logger.error('%s', e)
					exit(1)

				tavgk = 'tavg%d' % ((i + 1) % 2,)

				# async wrt. other streams & host, but not this stream.
				if i >= n_streams:
					stream.synchronize()
					tavg_unpinned.append(tavg[i % n_streams].copy())

				drv.memcpy_dtoh_async(tavg[i % n_streams], gpu_data[tavgk].ptr, stream=stream)

			# recover uncopied data from pinned buffer
			if nstep > n_streams:
				for i in range(nstep % n_streams, n_streams):
					stream.synchronize()
					tavg_unpinned.append(tavg[i].copy())

			for i in range(nstep % n_streams):
				stream.synchronize()
				tavg_unpinned.append(tavg[i].copy())

		except drv.LogicError as e:
			self.logger.error('%s. Check the number of states of the model or '
						 'GPU block shape settings blockdim.x/y %r, griddim %r.',
						 e, final_block_dim, final_grid_dim)
			exit(1)
		except drv.RuntimeError as e:
			self.logger.error('%s', e)
			exit(1)


		# self.logger.info('kernel finish..')
		# release pinned memory
		tavg = np.array(tavg_unpinned)

		# also release gpu_data
		self.release_gpumem(gpu_data)

		self.logger.info('kernel finished')
		return tavg

	def plot_output(self, tavg):
		plt.plot((tavg[:, self.args.plot_data, :, 0]), 'k', alpha=.2)
		plt.show()

	def write_output(self, tavg):
		tavg_file = open('tavg_data', 'wb')
		pickle.dump(tavg, tavg_file)
		tavg_file.close()

	def run_all(self):

		np.random.seed(79)

		tic = time.time()

		tavg0 = self.run_simulation()
		toc = time.time()
		elapsed = toc - tic

		if (self.args.validate == True):
			self.compare_with_ref(tavg0)

		self.plot_output(tavg0) if self.args.plot_data is not None else None
		self.write_output(tavg0) if self.args.write_data else None
		self.logger.info('Output shape (simsteps, states, bnodes, n_params) %s', tavg0.shape)
		self.logger.info('Finished CUDA simulation successfully in: {0:.3f}'.format(elapsed))
		self.logger.info('and in {0:.3f} M step/s'.format(
			1e-6 * self.args.n_time * self.n_inner_steps * self.n_work_items / elapsed))

		return tavg0


if __name__ == '__main__':

	driver_setup = Driver_Setup()
	tavgGPU = Driver_Execute(driver_setup).run_all()

	simtime = driver_setup.args.n_time
	# simtime = 10
	regions = driver_setup.args.n_regions
	g = 1.0
	# g = 0.0042
	s = 1.0
	dt = driver_setup.dt
	period = 1

	# generic model definition
	model = driver_setup.args.model.capitalize()+'T'

	# non generic model names
	# model = 'MontbrioT'
	# model = 'RwongwangT'
	# model = 'OscillatorT'
	# model = 'DumontGutkin'
	# model = 'MontbrioPazoRoxin'
	# model='Generic2dOscillator'
	(time, tavgCPU) = regularRun(simtime, g, s, dt, period).simulate_python(model)

	print('CPUshape', tavgCPU.shape)
	print('GPUshape', tavgGPU.shape)

	# check for deviation tolerance between GPU and CPU
	# for basic coupling and period = 1
	# using euler deterministic solver
	max_err = []
	x = 0
	for t in range(0, simtime):
		# print(t, 'tol:', np.max(np.abs(actual[t] - expected[t, :, :, 0])))
		# print(t, 'tol:', np.max(np.abs(tavgCPU[t,:,:,0], tavgGPU[t,:,:,0])))
		# print(t)
		# print('C', tavgCPU[t,:,:,0])
		# print('G', tavgGPU[t,:,:,0])
		# print(t, 'tol:', np.max(np.abs(tavgCPU[t,:,:,0] - tavgGPU[t,:,:,0])))
		np.testing.assert_allclose(tavgCPU[t, :, :, 0], tavgGPU[t, :, :, 0], 2e-5 * t * 2, 1e-5 * t * 2)


from tvb.simulator.lab import *

import numpy as np
import math

import matplotlib.pyplot as plt

from tvb.rateML.XML2model import RateML

class regularRun:

	def __init__(self, sim_length, g, s, dt, period, omega = 60, filename='connectivity_68.zip'):
	# def __init__(self, sim_length, g, s, dt, period, omega = 60, filename='paupau.zip'):
		self.sim_length = sim_length
		self.g = np.array([g])
		self.s = np.array([s])
		self.dt = dt
		self.period = period
		self.omega = omega * 2.0 * math.pi / 1e3
		(self.connectivity, self.coupling) = self.tvb_connectivity(filename)
		self.SC = self.connectivity.weights
		

	def tvb_connectivity(self, filename):
		white_matter = connectivity.Connectivity.from_file(source_file=filename)
		white_matter.configure()
		white_matter.speed = np.array([self.s])
		# white_matter_coupling = coupling.Linear(a=self.g)
		# white_matter_coupling = coupling.Scaling(a=self.g)
		white_matter_coupling = coupling.Coupling()
		return white_matter, white_matter_coupling
	
	def tvb_python_model(self, modelExec):
		# populations = models.Generic2dOscillator()	# original
		# populations = models.KuramotoT()				# generated
		# populations = models.OscillatorT()			# generated
		# populations = models.MontbrioT()				# generated
		# populations = models.RwongwangT()				# generated
		# populations = models.EpileptorT()				# generated
		model = 'models.' + modelExec + '()'
		populations = eval(model)
		populations.configure()
		populations.omega = np.array([self.omega])
		return populations


	def simulate_python(self, modelExec):
		# Initialize Model
		model = self.tvb_python_model(modelExec)
		# Initialize integrator
		# integrator = integrators.EulerDeterministic(dt=self.dt)
		# integrator = integrators.EulerStochastic(dt=1, noise=noise.Additive(nsig=np.array([1e-5])))
		integrator = integrators.EulerDeterministic(dt=self.dt)#, noise=noise.Additive(nsig=np.array([1e-5])))
		# Initialize Monitors
		monitorsen = (monitors.TemporalAverage(period=self.period))
		# Initialize Simulator
		sim = simulator.Simulator(model=model, connectivity=self.connectivity,
								  coupling=self.coupling,
								  integrator=integrator,
								  monitors=[monitorsen])
		sim.configure()
		sim.history.buffer[:] = 0.0
		sim.current_state[:] = 0.0
		print('shb', sim.history.buffer.shape)  # ('n_time', 'n_cvar', 'n_node', 'n_mode')

		(time, data) = sim.run(simulation_length=self.sim_length)[0]

		# pad some zeros to make it equivalent to GPU for comparison
		# CPU is now +1 timestep longer
		data = np.insert(data, 0, 0, axis=0)
		data = data[:-1]

		# print('ds',data.shape)
		plt.plot((data[:, 0, :, 0]), 'k', alpha=.2)
		plt.show()

		return (time, data)

if __name__ == '__main__':

	# model_filename = 'Oscillator'
	# language='python'
	#
	# RateML(model_filename, language)

	simtime = 20
	g = 1
	# g = 0.0042
	s = 1.0
	dt = .1
	period = 1

	model = 'OscillatorT'
	# model = 'MontbrioPazoRoxin'
	# model='Generic2dOscillator'
	(time, data) = regularRun(simtime, g, s, dt, period).simulate_python(model)


from tvb.simulator.models.base import Model, ModelNumbaDfun
import numpy
from numpy import *
from numba import guvectorize, float64
from tvb.basic.neotraits.api import NArray, Final, List, Range

class MontbrioT(ModelNumbaDfun):
        
    tau = NArray(
        label=":math:`tau`",
        default=numpy.array([1.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.01),
        doc=""""""
    )    
        
    I = NArray(
        label=":math:`I`",
        default=numpy.array([0.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.01),
        doc=""""""
    )    
        
    Delta = NArray(
        label=":math:`Delta`",
        default=numpy.array([0.7]),
        domain=Range(lo=0.0, hi=10.0, step=0.01),
        doc=""""""
    )    
        
    J = NArray(
        label=":math:`J`",
        default=numpy.array([14.5]),
        domain=Range(lo=-25.0, hi=25.0, step=0.0001),
        doc=""""""
    )    
        
    eta = NArray(
        label=":math:`eta`",
        default=numpy.array([-4.6]),
        domain=Range(lo=-10.0, hi=10.0, step=0.0001),
        doc=""""""
    )    
        
    Gamma = NArray(
        label=":math:`Gamma`",
        default=numpy.array([5.0]),
        domain=Range(lo=0., hi=10.0, step=0.1),
        doc=""""""
    )    
        
    cr = NArray(
        label=":math:`cr`",
        default=numpy.array([1.0]),
        domain=Range(lo=0., hi=1, step=0.1),
        doc=""""""
    )    
        
    cv = NArray(
        label=":math:`cv`",
        default=numpy.array([1.0]),
        domain=Range(lo=0., hi=1, step=0.1),
        doc=""""""
    )    

    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={"r": numpy.array([0.0, -2.0]), 
				 "V": numpy.array([-2.0, 1.5])},
        doc="""state variables"""
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={"r": numpy.array([0.0, inf]), 
				 },
    )
    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=('r', 'V', ),
        default=('r', 'V', ),

        doc="Variables to monitor"
    )

    state_variables = ['r', 'V']

    _nvar = 2
    cvar = numpy.array([0,1,], dtype = numpy.int32)

    def dfun(self, vw, c, local_coupling=0.0):
        vw_ = vw.reshape(vw.shape[:-1]).T
        c_ = c.reshape(c.shape[:-1]).T
        deriv = _numba_dfun_MontbrioT(vw_, c_, self.tau, self.I, self.Delta, self.J, self.eta, self.Gamma, self.cr, self.cv, local_coupling)

        return deriv.T[..., numpy.newaxis]

@guvectorize([(float64[:], float64[:], float64, float64, float64, float64, float64, float64, float64, float64, float64, float64[:])], '(n),(m)' + ',()'*9 + '->(n)', nopython=True)
def _numba_dfun_MontbrioT(vw, coupling, tau, I, Delta, J, eta, Gamma, cr, cv, local_coupling, dx):
    "Gufunc for MontbrioT model equations."

    # long-range coupling
    c_pop0 = coupling[0]
    c_pop1 = coupling[1]

    r = vw[0]
    V = vw[1]


    dx[0] = 1/tau * (Delta / (pi * tau) + 2 * V * r)
    dx[1] = 1/tau * (V**2 - pi**2 * tau**2 * r**2 + eta + J * tau * r + I + cr * c_pop0 + cv * c_pop1)
    



from tvb.simulator.models.base import Model, ModelNumbaDfun
import numpy
from numpy import *
from numba import guvectorize, float64
from tvb.basic.neotraits.api import NArray, Final, List, Range

class KuramotoT(ModelNumbaDfun):
        
    omega = NArray(
        label=":math:`omega`",
        default=numpy.array([60.0 * 2.0 * 3.1415927 / 1e3]),
        doc=""""""
    )    

    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={"V": numpy.array([0.0])},
        doc="""state variables"""
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={"V": numpy.array([-2, 1])},
    )
    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=('V', ),
        default=('V', ),
        doc="Variables to monitor"
    )

    state_variables = ['V']

    _nvar = 1
    cvar = numpy.array([0,], dtype = numpy.int32)

    def dfun(self, vw, c, local_coupling=0.0):
        vw_ = vw.reshape(vw.shape[:-1]).T
        c_ = c.reshape(c.shape[:-1]).T
        deriv = _numba_dfun_KuramotoT(vw_, c_, self.omega, local_coupling)

        return deriv.T[..., numpy.newaxis]

@guvectorize([(float64[:], float64[:], float64, float64, float64[:])], '(n),(m)' + ',()'*2 + '->(n)', nopython=True)
def _numba_dfun_KuramotoT(vw, coupling, omega, local_coupling, dx):
    "Gufunc for KuramotoT model equations."

    # long-range coupling
    c_pop0 = coupling[0]

    V = vw[0]


    dx[0] = omega + c_pop1
    
